"""Prometheus Grok Exporter module."""
import collections
import json
import os
import socket
import subprocess
from pathlib import Path

from charmhelpers.core import hookenv, host, unitdata
from charmhelpers.core.templating import render

import charms.promreg
from charms.layer import snap

import yaml


GROK_PATTERNS_CONFIG_PATH = "files/grok-patterns.yaml"
METRICS_MAPPING_CONFIG_PATH = "files/metrics-mapping.yaml"
DASHBOARD_PATH = "files/grafana-dashboards"


def is_valid_ip(listen_ip):
    """Check if listen_ip is configured on the unit as private or public address."""
    if listen_ip == hookenv.unit_get(
        "private-address"
    ) or listen_ip == hookenv.unit_get("public-address"):
        return True
    return False


def known_values(listen_ip):
    """Return a valid ip.

    Return 0.0.0.0 if listen ip is equal to 0.0.0.0 or all,
    return 127.0.0.1 if listen_ip is equal to 127.0.0.1 or localhost,
    return False otherwise.
    """
    default_value, localhost = "0.0.0.0", "127.0.0.1"
    values = collections.defaultdict(lambda: False)
    values.update(
        {
            "0.0.0.0": default_value,
            "all": default_value,
            "localhost": localhost,
            "127.0.0.1": localhost,
        }
    )
    return values[listen_ip]


def choose_bind_ip(listen_ip):
    """Return the listen ip for the prometheus grok-exporter.

    Default value is 0.0.0.0.
    """
    if known_values(listen_ip):
        return known_values(listen_ip)

    if not is_valid_ip(listen_ip):
        hookenv.log(
            "listen_address not valid. Using default '0.0.0.0' instead",
            level=hookenv.INFO,
        )
        return "0.0.0.0"

    return listen_ip


def parse_extra_metric(metric, optional_keyword="cumulative"):
    """Parse extra parameters for not counter metrics.

    :param metric: The raw metric coming from the parse_metric function.
    :type metric: Dict[str, str]
    :param optional_keyword: the optional keyword depending on the metric type.
    :type optional_keyword: str
    :returns: A Dictionary with value and the optional keyword if present or False
          if the metric is not valid.
    :rtype: Union[bool, Dict[str, str]]
    """
    if "value" not in metric:
        return False
    parsed_metric = {"value": metric["value"]}

    if optional_keyword in metric:
        parsed_metric[optional_keyword] = metric[optional_keyword]

    return parsed_metric


def parse_common_metric(metric):
    """Validate and parse the common options of a metric.

    :param metric: The raw metric coming from the parse_metric function.
    :type metric: Dict[str, str]
    :returns: A Dictionary with a name, match, and several optional fields that can
    describe any type of metric or False if the metric is not valid.
    :rtype: Union[bool, Dict[str, str]]
    """
    if "name" not in metric or "match" not in metric or "help" not in metric:
        return False

    if "path" in metric and "paths" in metric:
        return False

    parsed_metric = {
        "name": metric["name"],
        "match": metric["match"],
        "help": metric["help"],
    }

    if "path" in metric:
        parsed_metric["paths"] = [metric["path"]]

    if "paths" in metric:
        parsed_metric["paths"] = metric["paths"]

    if "labels" in metric:
        parsed_metric["labels"] = metric["labels"]

    return parsed_metric


def parse_metric(metric):
    """Validate and parse a metric.

    :param metric: The raw metric coming from the charm configuration option
    :type metric: Dict[str, str]
    :returns: A Dictionary with type, name, match, and several optional fields that
          describe a metric or False if the metric is not valid.
    :rtype: Union[bool, Dict[str, str]]
    """
    if "type" not in metric or metric["type"] not in (
        "counter",
        "gauge",
        "histogram",
        "summary",
    ):
        return False

    parsed_metric = {"type": metric["type"]}

    common = parse_common_metric(metric)

    if not common:
        return False

    parsed_metric.update(common)

    metric_optional_keywords = {
        "counter": None,
        "gauge": "cumulative",
        "histogram": "buckets",
        "summary": "quantiles",
    }

    metric_optional_keyword = metric_optional_keywords[metric["type"]]

    if metric_optional_keyword:
        extra_metric = parse_extra_metric(
            metric, optional_keyword=metric_optional_keyword
        )
        if not extra_metric:
            return False
        parsed_metric.update(extra_metric)

    if "labels" not in parsed_metric:
        parsed_metric["labels"] = {}

    if "host" not in parsed_metric["labels"]:
        parsed_metric["labels"]["host"] = socket.gethostname()

    return parsed_metric


class PrometheusGrokExporterHelper:
    """Prometheus Grok Exporter Helper."""

    def __init__(self, grafana_endpoint=None):
        """Initiliatize the helper."""
        self.charm_config = hookenv.config()
        self.metrics_yaml = None
        self.unitdata_global_paths_key = "prometheus_grok_exporter.global_paths"
        self.grafana_endpoint = grafana_endpoint

    @property
    def snap_channel(self):
        """Return the snap channel. Default: stable."""
        return self.charm_config.get("snap-channel")

    @property
    def global_paths(self):
        """Return the paths of the log files to which all the metrics apply.

        Default: empty list.
        """
        global_paths = self.charm_config.get("global-paths")

        if not global_paths:
            global_paths = []
        else:
            global_paths = global_paths.split(",")

        # Add auto discovered global paths
        global_paths += self.available_global_paths_from_mapping()
        return set(global_paths)

    @property
    def fail_on_missing_logfile(self):
        """Return true if exporter should fail if log files to monitor don't exist."""
        return self.charm_config.get("fail-on-missing-logfile")

    @property
    def grok_patterns_path(self):
        """Return where the grok patters are stored."""
        return self.charm_config.get("grok-patterns-path")

    @property
    def grok_exporter_config_path(self):
        """Return where the grok-exporter config file is stored."""
        return "/var/snap/{}/current/etc/config.yml".format(self.snap)

    @property
    def snap(self):
        """Return the snap name of the grok-exporter."""
        return "prometheus-grok-exporter"

    @property
    def svc_name(self):
        """Return the service name of the grok-exporter."""
        return "snap.{}.grok-exporter".format(self.snap)

    @property
    def listen_address(self):
        """Return the address grok-exporter will listen on to expose the metrics."""
        return self.charm_config.get("listen-address")

    @property
    def port(self):
        """Return the port grok-exporter will listen on to expose the metrics."""
        return self.charm_config.get("port")

    @property
    def metrics(self):
        """Return the metrics configured on the charm in raw format."""
        return self.charm_config.get("metrics")

    @property
    def validated_listen_address(self):
        """Return the validated listen address using the chose_bind_ip func."""
        return choose_bind_ip(self.listen_address)

    @property
    def parsed_metrics(self):
        """Return the parsed metrics as a list."""
        metrics = []
        self.metrics_yaml = yaml.safe_load(self.metrics)
        if self.metrics_yaml and "metrics" in self.metrics_yaml:
            metrics += list(
                map(
                    parse_metric,
                    list(filter(parse_metric, self.metrics_yaml["metrics"])),
                )
            )
        metrics += self.available_metrics_from_mapping()
        return metrics

    @property
    def enable_autodiscovery(self):
        """Return true if metrics autodiscovery is enabled."""
        return self.charm_config.get("enable-autodiscovery")

    @property
    def grafana_dashboards(self):
        """Return the grafana dashboards."""
        dash_dir = Path(DASHBOARD_PATH)
        return dash_dir.glob("*.json")

    def register_grafana_dashboards(self):
        """Register grafana dashboards."""
        if not self.grafana_endpoint:
            return

        for dash_file in self.grafana_dashboards:
            dashboard = dash_file.read_text()
            self.grafana_endpoint.register_dashboard(
                dash_file.stem, json.loads(dashboard)
            )
            hookenv.log("Pushed grafana dashboard {}".format(dash_file))

    def available_global_paths_from_mapping(self):
        """Return a list of existing files that are defined in the metrics mapping."""
        available_global_paths = []
        with open(METRICS_MAPPING_CONFIG_PATH, "r") as metrics_mapping_file:
            metrics_mapping = yaml.safe_load(metrics_mapping_file)
            metrics = metrics_mapping["metrics"]
            for metric in metrics:
                if (
                    os.path.exists(metrics[metric]["path"])
                    and self.enable_autodiscovery
                ):
                    available_global_paths.append(metrics[metric]["path"])
        return available_global_paths

    def available_metrics_from_mapping(self):
        """Return a list of metrics that are defined in the metrics mapping.

        Only metrics that refer to log files that exists are returned.
        """
        available_metrics = []
        with open(METRICS_MAPPING_CONFIG_PATH, "r") as metrics_mapping_file:
            metrics_mapping = yaml.safe_load(metrics_mapping_file)
            metrics = metrics_mapping["metrics"]
            for metric in metrics:
                if (
                    os.path.exists(metrics[metric]["path"])
                    and self.enable_autodiscovery
                ):
                    metrics[metric]["name"] = metric
                    available_metrics.append(parse_metric(metrics[metric]))
        return available_metrics

    def get_prometheus_client_ip(self, interface="prometheus-client"):
        """Return the ip to register on prometheus."""
        if known_values(self.validated_listen_address):
            return hookenv.network_get_primary_address(interface)
        return self.validated_listen_address

    def are_metrics_valid(self):
        """Validate the metrics."""
        try:
            self.metrics_yaml = yaml.safe_load(self.metrics)
            parsed_metrics = []
            if self.metrics_yaml and "metrics" in self.metrics_yaml:
                parsed_metrics = list(
                    filter(parse_metric, self.metrics_yaml["metrics"])
                )
                if len(parsed_metrics) != len(self.metrics_yaml["metrics"]):
                    return False
        except yaml.YAMLError:
            hookenv.log(
                "Metrics is not a valid YAML. Check configuration.", hookenv.INFO
            )
            return False

        if len(self.available_metrics_from_mapping()) > 0 or (
            len(parsed_metrics) != 0
            and len(parsed_metrics) == len(self.metrics_yaml["metrics"])
        ):
            return True

        return False

    def open_port(self):
        """Open the port the grok-exporter will listen on to expose the metrics.

        If the port changes, the old port is closed.
        """
        unitdata_key = "prometheus_grok_exporter.port"
        kv = unitdata.kv()
        existing_port = kv.get(unitdata_key)
        if existing_port != self.port:
            hookenv.open_port(self.port)
            if existing_port:  # Dont try to close non existing ports
                hookenv.close_port(existing_port)
                charms.promreg.deregister(None, existing_port)
            kv.set(unitdata_key, self.port)

        # Register the port, even if this is a re-registration
        charms.promreg.register(None, self.port)

    def install_snap(self):
        """Install the snap and connect the log-observe interface."""
        snap.install(self.snap, channel=self.snap_channel, force_dangerous=False)

        # TODO: remove 3 calls below when we get autoconnect assertion
        subprocess.check_call(["snap", "connect", "{}:log-observe".format(self.snap)])
        subprocess.check_call(["snap", "disable", self.snap])
        subprocess.check_call(["snap", "enable", self.snap])

    def uninstall_snap(self):
        """Remove the snap."""
        snap.remove(self.snap)

    def get_grok_patterns(self):
        """Parse grok-patterns.yaml and return the grok-patterns in it."""
        with open(GROK_PATTERNS_CONFIG_PATH, "r") as grok_pattern_file:
            return yaml.safe_load(grok_pattern_file)["grok_patterns"]

    def get_context(self):
        """Get the context to render the grok-exporter config file."""
        return {
            "fail_on_missing_logfile": self.fail_on_missing_logfile,
            "global_paths": self.global_paths,
            "grok_patterns_path": self.grok_patterns_path,
            "metrics": self.parsed_metrics,
            "listen_ip": self.validated_listen_address,
            "listen_port": self.port,
            "grok_patterns": self.get_grok_patterns(),
        }

    def write_grok_config(self):
        """Write the grok-exporter config file if the metrics are valid.

        If the metrics are not valid, it doesn't do anything.
        """
        if self.are_metrics_valid():
            ctx = self.get_context()
            render(
                source="grok-exporter.yml.j2",
                templates_dir="templates",
                target=self.grok_exporter_config_path,
                context=ctx,
            )
            kv = unitdata.kv()
            kv.set(self.unitdata_global_paths_key, ",".join(ctx["global_paths"]))

    def is_service_running(self):
        """Return true if the grok-exporter service is running."""
        return host.service_running(self.svc_name)

    def start_service(self, restart=False):
        """Start or restart the grok-exporter service."""
        if restart:
            return host.service_restart(self.svc_name)
        return host.service_start(self.svc_name)

    def check_status(self):
        """Check the status of the service."""
        # When checking the status check if any path from metrics-mapping
        # not yet included in the configuration now exists.
        # If you find a new one rewrite the configuration.
        kv = unitdata.kv()
        global_paths_from_unitdata = kv.get(self.unitdata_global_paths_key, "")
        renew_config = False
        for path in self.global_paths:
            if path not in global_paths_from_unitdata:
                renew_config = True

        if renew_config:
            self.write_grok_config()
            self.start_service(restart=self.is_service_running())

        if not self.are_metrics_valid():
            return "metric_not_valid"
        if not self.global_paths or len(self.global_paths) == 0:
            return "no_global_paths"
        if not self.is_service_running():
            return "service_not_running"

        return "service_running"
