"""Reactive charm hooks."""
import os
import time

from charmhelpers.contrib.charmsupport import nrpe
from charmhelpers.core import hookenv

from charms.reactive import (
    endpoint_from_flag,
    hook,
    remove_state,
    set_state,
    when,
    when_any,
    when_not,
)

from lib_prometheus_grok_exporter import PrometheusGrokExporterHelper


charm_statues = {
    "metrics_not_valid": ("blocked", "Metrics are not valid. Check the configurations"),
    "no_global_paths": (
        "blocked",
        "No global paths found. Configure 'global-paths' charm option",
    ),
    "service_not_running": (
        "blocked",
        "Service is not running. Check that files defined in 'global-paths' exists "
        "or manually check the service running on the unit the command "
        "sudo snap run prometheus-grok-exporter.grok-exporter",
    ),
    "service_running": ("active", "Unit is ready"),
}


@hook("upgrade-charm")
def upgrade_charm():
    """Run reconfiguration of the charm."""
    set_state("grok-exporter.do-reconfig")


@hook("update-status")
def update_status():
    """Update the status of the unit checking if grok-exporter is running."""
    prom_grok_exporter_helper = PrometheusGrokExporterHelper()
    status = charm_statues[prom_grok_exporter_helper.check_status()]
    hookenv.status_set(status[0], status[1])


@when("juju-info.connected")
@when_not("grok-exporter.installed")
def install_packages():
    """Install the snap."""
    hookenv.status_set("maintenance", "Installing software")

    prom_grok_exporter_helper = PrometheusGrokExporterHelper()
    prom_grok_exporter_helper.install_snap()

    set_state("grok-exporter.installed")
    set_state("grok-exporter.do-reconfig")


@when("grok-exporter.installed")
@when_any("grok-exporter.do-reconfig", "config.changed")
def write_config():
    """Write the grok-exporter configuration."""
    prom_grok_exporter_helper = PrometheusGrokExporterHelper()

    if not prom_grok_exporter_helper.are_metrics_valid():
        status = charm_statues["metrics_not_valid"]
        hookenv.status_set(status[0], status[1])
    elif (
        not prom_grok_exporter_helper.global_paths
        or len(prom_grok_exporter_helper.global_paths) == 0
    ):
        status = charm_statues["no_global_paths"]
        hookenv.status_set(status[0], status[1])
    else:
        prom_grok_exporter_helper.write_grok_config()
        prom_grok_exporter_helper.open_port()
        remove_state("grok-exporter.nrpe_configured")
        remove_state("grok-exporter.do-reconfig")
        remove_state("grok-exporter.prometheus_relation_configured")
        set_state("grok-exporter.do-restart")
        set_state("grok-exporter.register-dashboards")


@when("config.changed.snap_channel")
def snap_channel_changed():
    """Remove the state grok.exporter.installed if the snap channel changes."""
    remove_state("grok-exporter.installed")


@when("grok-exporter.do-restart")
def restart_grok_exporter():
    """Restart the grok-exporter service."""
    prom_grok_exporter_helper = PrometheusGrokExporterHelper()
    if not prom_grok_exporter_helper.is_service_running():
        hookenv.log("Starting {}...".format(prom_grok_exporter_helper.svc_name))
        prom_grok_exporter_helper.start_service()
    else:
        hookenv.log(
            "Restarting {}, config file changed...".format(
                prom_grok_exporter_helper.svc_name
            )
        )
        prom_grok_exporter_helper.start_service(restart=True)

    set_state("grok-exporter.started")
    remove_state("grok-exporter.do-restart")
    time.sleep(2)  # Wait for restart to happen
    status = charm_statues[prom_grok_exporter_helper.check_status()]
    hookenv.status_set(status[0], status[1])
    hookenv.log(status[1])


# Relations
@when("grok-exporter.started")
@when("prometheus-client.available")
@when_not("grok-exporter.prometheus_relation_configured")
def configure_prometheus_relation(target):
    """Register the exporter to prometheus."""
    prom_grok_exporter_helper = PrometheusGrokExporterHelper()
    hostname = prom_grok_exporter_helper.get_prometheus_client_ip()
    prom_grok_exporter_helper = PrometheusGrokExporterHelper()
    target.configure(prom_grok_exporter_helper.port, hostname=hostname)
    set_state("grok-exporter.prometheus_relation_configured")


@when("nrpe-external-master.available")
@when_not("grok-exporter.nrpe_configured")
def update_nrpe_config(svc):
    """Add the nrpe check http to verify that grok-exporter is healthy."""
    if not os.path.exists("/var/lib/nagios"):
        hookenv.status_set("blocked", "Waiting for nrpe package installation")
        return

    hostname = nrpe.get_nagios_hostname()
    nrpe_setup = nrpe.NRPE(hostname=hostname)
    prom_grok_exporter_helper = PrometheusGrokExporterHelper()
    port = prom_grok_exporter_helper.port
    nrpe_setup.add_check(
        "prometheus_grok_exporter_http",
        "Prometheus Grok Exporter HTTP check",
        "check_http -I 127.0.0.1 -p {} -u /metrics".format(port),
    )
    nrpe_setup.write()
    set_state("grok-exporter.nrpe_configured")


@when("nrpe-external-master.changed")
def nrpe_changed():
    """Trigger nrpe update."""
    remove_state("grok-exporter.nrpe_configured")


@when("prometheus-client.changed")
def prometheus_changed():
    """Trigger prometheus update."""
    remove_state("grok-exporter.prometheus_relation_configured")


# Clean up
@when("grok-exporter.installed")
@when_not("juju-info.available")
def remove_grok_exporter():
    """Uninstall the snap."""
    prom_grok_exporter_helper = PrometheusGrokExporterHelper()
    prom_grok_exporter_helper.uninstall_snap()
    remove_state("grok-exporter.installed")
    remove_state("grok-exporter.started")


@when("grok-exporter.nrpe_configured")
@when_not("nrpe-external-master.available")
def remove_nrpe_check():
    """Remove the nrpe check."""
    hostname = nrpe.get_nagios_hostname()
    nrpe_setup = nrpe.NRPE(hostname=hostname)
    nrpe_setup.remove_check(shortname="prometheus_grok_exporter_http")
    remove_state("grok-exporter.nrpe_configured")


@when(
    "leadership.is_leader",
    "endpoint.dashboards.joined",
    "grok-exporter.register-dashboards",
)
def register_grafana_dashboards():
    """Register grafana dashboards."""
    grafana_endpoint = endpoint_from_flag("endpoint.dashboards.joined")
    prom_grok_exporter_helper = PrometheusGrokExporterHelper(
        grafana_endpoint=grafana_endpoint
    )
    prom_grok_exporter_helper.register_grafana_dashboards()
    remove_state("grok-exporter.register-dashboards")
