"""Encapsulate prometheus-grok-exporter testing."""

import json
import logging
import time
import unittest

import requests

import zaza.model as model


CURL_TIMEOUT = 180
RUN_ACTION_TIMEOUT = 180
DEFAULT_API_PORT = "9144"
DEFAULT_API_URL = "/metrics"


class BasePromGrokExporterTest(unittest.TestCase):
    """Base for Prometheus-grok-exporter charm tests."""

    @classmethod
    def setUpClass(cls):
        """Set up tests."""
        super(BasePromGrokExporterTest, cls).setUpClass()
        cls.model_name = model.get_juju_model()
        cls.application_name = "prometheus-grok-exporter"
        cls.lead_unit_name = model.get_lead_unit_name(
            cls.application_name, model_name=cls.model_name
        )
        cls.units = model.get_units(cls.application_name, model_name=cls.model_name)

        cls.prom_grok_exporter_ip = model.get_app_ips(cls.application_name)[0]


class CharmOperationTest(BasePromGrokExporterTest):
    """Verify operations."""

    def test_01_api_ready(self):
        """Verify if the API is ready.

        Curl the api endpoint.
        We'll retry until the CURL_TIMEOUT.
        """
        curl_command = "curl http://localhost:{}/metrics".format(DEFAULT_API_PORT)
        timeout = time.time() + CURL_TIMEOUT
        while time.time() < timeout:
            response = model.run_on_unit(self.lead_unit_name, curl_command)
            if response["Code"] == "0":
                return
            logging.warning(
                "Unexpected curl response: {}. Retrying in 30s.".format(response)
            )
            time.sleep(30)

        # we didn't get rc=0 in the allowed time, fail the test
        self.fail(
            "Prometheus-grok-exporter didn't respond to the command \n"
            "'{curl_command}' as expected.\n"
            "Result: {result}".format(curl_command=curl_command, result=response)
        )

    def test_02_nrpe_http_check(self):
        """Verify nrpe check exists."""
        expected_nrpe_check = (
            "command[check_prometheus_grok_exporter_http]={} -I "
            "127.0.0.1 -p {} -u {}".format(
                "/usr/lib/nagios/" "plugins/check_http",
                DEFAULT_API_PORT,
                DEFAULT_API_URL,
            )
        )
        logging.debug(
            "Verify the nrpe check is created and has the required content..."
        )
        cmd = "cat /etc/nagios/nrpe.d/check_prometheus_grok_exporter_http.cfg"
        result = model.run_on_unit(self.lead_unit_name, cmd)
        code = result.get("Code")
        if code != "0":
            raise model.CommandRunFailed(cmd, result)
        content = result.get("Stdout")
        self.assertTrue(expected_nrpe_check in content)

    def test_03_default_grafana_dashboard(self):
        """Verify the default dashboard is imported in grafana."""
        if not model.get_application("grafana"):
            # if grafana not there the test is not valid
            return self.assertTrue(True, "Grafana is not deployed. Test skipped.")

        action = model.run_action_on_leader("grafana", "get-admin-password")

        timeout = time.time() + RUN_ACTION_TIMEOUT
        while time.time() < timeout:
            if action.status == "completed":
                break

        grafana_password = action.data.get("results", {}).get("password", "")
        grafana_port = model.get_application_config("grafana").get("port").get("value")
        grafana_host = model.get_lead_unit_ip("grafana")
        req = requests.get(
            "http://{host}:{port}"
            "/api/search?dashboardIds".format(host=grafana_host, port=grafana_port),
            auth=("admin", grafana_password),
        )
        # Collect imported dashboards titles
        titles = []
        if req.status_code == 200:
            dashboards = json.loads(req.text)
            for dash in dashboards:
                titles.append(dash.get("title", ""))

        self.assertIn("[juju-prometheus-grok-exporter] Grok Exporter", titles)
