#!/usr/bin/python3
"""Test lib_prometheus_grok_exporter module."""

import mock

import pytest


class TestLib:
    """Test the lib_prometheus_grok_exporter class."""

    def test_pytest(self):
        """Assert pytest is used."""
        assert True

    def test_prometheusgrokexporter(self, prometheusgrokexporter):
        """See if the helper fixture works to load charm configs."""
        assert isinstance(prometheusgrokexporter.charm_config, dict)

    @pytest.mark.parametrize(
        "listen_address",
        ["0.0.0.0", "127.0.0.1", "localhost", "all", "10.0.0.10", "1.1.1.1"],
    )
    @mock.patch("lib_prometheus_grok_exporter.hookenv.unit_get")
    @mock.patch("lib_prometheus_grok_exporter.hookenv.network_get_primary_address")
    def test_get_prometheus_client_ip(
        self, get_primary_address, unit_get, prometheusgrokexporter, listen_address
    ):
        """Test get_prometheus_client_ip."""
        unit_ip = "10.0.0.10"
        get_primary_address.return_value = unit_ip
        unit_get.return_value = unit_ip

        # Test default value
        prometheusgrokexporter.charm_config["listen-address"] = listen_address
        assert prometheusgrokexporter.get_prometheus_client_ip() == unit_ip

    def test_valid_metrics(self, prometheusgrokexporter):
        """Test metrics validation and parsing."""
        # Test valid yaml
        metrics_valid_yaml = """
        metrics:
        - type: counter
          name: syslog_msgs_errors_total
          help: Total number of syslog msgs with errors, by program (ie daemon)
          match: "%{SYSLOGBASE}.*[Ee]rrors?(?!${WORD})"
          labels:
            program: "{{.program}}"
          path: /var/log/syslog
        - type: gauge
          name: grok_example_values
          help: Example gauge metric with labels.
          match: "%{DATE} %{TIME} %{USER:user} %{NUMBER:val}"
          value: "{{.val}}"
          cumulative: false
          labels:
            user: "{{.user}}"
        - type: histogram
          name: grok_example_values_1
          help: Example histogram metric with labels.
          match: "%{DATE} %{TIME} %{USER:user} %{NUMBER:val}"
          value: "{{.val}}"
          buckets: [1, 2, 3]
          labels:
            user: "{{.user}}"
        - type: summary
          name: grok_example_values_2
          help: Summary metric with labels.
          match: "%{DATE} %{TIME} %{USER:user} %{NUMBER:val}"
          value: "{{.val}}"
          quantiles: {0.5: 0.05, 0.9: 0.01, 0.99: 0.001}
          labels:
            user: "{{.user}}"
          paths:
          - /var/log/syslog
          - /var/log/kern.log
        """
        prometheusgrokexporter.charm_config["metrics"] = metrics_valid_yaml
        assert prometheusgrokexporter.are_metrics_valid()
        assert len(prometheusgrokexporter.parsed_metrics) == 4

    @mock.patch("lib_prometheus_grok_exporter.os.path.exists")
    def test_invalid_yaml_metrics(self, file_exist, prometheusgrokexporter):
        """Test invalid yaml."""
        metrics_invalid_yaml = """
        Not a = valid yaml file.
        """
        file_exist.return_value = False
        prometheusgrokexporter.charm_config["metrics"] = metrics_invalid_yaml

        assert not prometheusgrokexporter.are_metrics_valid()
        assert len(prometheusgrokexporter.parsed_metrics) == 0

    @pytest.mark.parametrize(
        "invalid_metric",
        [
            """
        metrics: # missing name
        - type: counter
          help: Total number of syslog msgs with errors, by program (ie daemon)
          match: "%{SYSLOGBASE}.*[Ee]rrors?(?!${WORD})"
          labels:
            program: "{{.program}}"
          path: /var/log/syslog
        """,
            """
        metrics: # contains both path and paths
        - type: counter
          help: Total number of syslog msgs with errors, by program (ie daemon)
          name: test
          match: "%{SYSLOGBASE}.*[Ee]rrors?(?!${WORD})"
          labels:
            program: "{{.program}}"
          path: /var/log/syslog
          paths:
              - /var/log/kern.log
        """,
            """
        metrics: # missing value
        - type: summary
          name: grok_example_values_2
          help: Summary metric with labels.
          match: "%{DATE} %{TIME} %{USER:user} %{NUMBER:val}"
          quantiles: {0.5: 0.05, 0.9: 0.01, 0.99: 0.001}
          labels:
            user: "{{.user}}"
        """,
            """
        metrics: # wrong type
        - type: wrong_type
          name: grok_example_values_2
          help: Summary metric with labels.
          match: "%{DATE} %{TIME} %{USER:user} %{NUMBER:val}"
          quantiles: {0.5: 0.05, 0.9: 0.01, 0.99: 0.001}
          labels:
            user: "{{.user}}"
        """,
        ],
    )
    @mock.patch("lib_prometheus_grok_exporter.os.path.exists")
    def test_invalid_metrics(self, file_exists, prometheusgrokexporter, invalid_metric):
        """Test invalid metrics."""
        file_exists.return_value = False
        prometheusgrokexporter.charm_config["metrics"] = invalid_metric
        assert not prometheusgrokexporter.are_metrics_valid()
        assert len(prometheusgrokexporter.parsed_metrics) == 0

    @mock.patch("lib_prometheus_grok_exporter.charms")
    @mock.patch("lib_prometheus_grok_exporter.hookenv.open_port")
    def test_open_port(self, open_port, charms, prometheusgrokexporter):
        """Test open port for the exporter."""
        prometheusgrokexporter.charm_config["port"] = 9144

        class PromReg:
            @classmethod
            def register(cls, target, port):
                return True

        charms.promreg.return_value = PromReg
        prometheusgrokexporter.open_port()
        open_port.assert_called_with(9144)
        charms.promreg.register.assert_called_with(None, 9144)

    @mock.patch("lib_prometheus_grok_exporter.subprocess.check_call")
    @mock.patch("lib_prometheus_grok_exporter.snap.install")
    def install_snap(self, snap_install, subprocess_check_call):
        """Test snap installation."""
        snap_install.assert_called_with(
            "prometheus-grok-exporter", channel="stable", force_dangerous=False
        )
        subprocess_check_call.assert_called_with(
            ["snap", "connect", "prometheus-grok-exporter:log-observe"]
        )
        subprocess_check_call.assert_called_with(
            ["snap", "disable", "prometheus-grok-exporter"]
        )
        subprocess_check_call.assert_called_with(
            ["snap", "enable", "prometheus-grok-exporter"]
        )

    def test_default_config(self, prometheusgrokexporter):
        """Test default configuration is returning the expected values."""
        assert prometheusgrokexporter.snap_channel == "stable"
        assert prometheusgrokexporter.global_paths == set()
        assert prometheusgrokexporter.fail_on_missing_logfile
        assert (
            prometheusgrokexporter.grok_patterns_path
            == "/var/snap/prometheus-grok-exporter"
            "/current/logstash-patterns-core/patterns"
        )
        assert prometheusgrokexporter.listen_address == "0.0.0.0"
        assert (
            prometheusgrokexporter.svc_name
            == "snap.prometheus-grok-exporter.grok-exporter"
        )
        assert prometheusgrokexporter.port == 9144
        assert prometheusgrokexporter.snap == "prometheus-grok-exporter"
        assert (
            prometheusgrokexporter.grok_exporter_config_path
            == "/var/snap/prometheus-grok-exporter/current/etc/config.yml"
        )
        assert prometheusgrokexporter.validated_listen_address == "0.0.0.0"
        assert prometheusgrokexporter.charm_config["nagios_context"] == "juju"
        assert prometheusgrokexporter.charm_config["nagios_servicegroups"] == "juju"

    def test_get_context(self, prometheusgrokexporter):
        """Test get_context for default configuration."""

        def get_grok_patterns():
            return {}

        prometheusgrokexporter.get_grok_patterns = get_grok_patterns

        return_value = {
            "fail_on_missing_logfile": True,
            "global_paths": set(),
            "grok_patterns_path": "/var/snap/prometheus-grok-exporter"
            "/current/logstash-patterns-core/patterns",
            "metrics": [],
            "listen_ip": "0.0.0.0",
            "listen_port": 9144,
            "grok_patterns": {},
        }

        assert prometheusgrokexporter.get_context() == return_value

    @mock.patch("lib_prometheus_grok_exporter.socket.gethostname")
    @mock.patch("lib_prometheus_grok_exporter.render")
    def test_write_grok_config(self, render, get_hostname, prometheusgrokexporter):
        """Test writing grok exporter configuration."""

        def get_grok_patterns():
            return {}

        prometheusgrokexporter.get_grok_patterns = get_grok_patterns

        def available_metrics_from_mapping():
            return []

        prometheusgrokexporter.available_metrics_from_mapping = (
            available_metrics_from_mapping
        )

        def available_global_paths_from_mapping():
            return []

        prometheusgrokexporter.available_global_paths_from_mapping = (
            available_global_paths_from_mapping
        )

        hostname = "myhostname"
        get_hostname.return_value = hostname

        metrics_valid_yaml = """
        metrics:
        - type: counter
          name: syslog_msgs_errors_total
          help: Total number of syslog msgs with errors, by program (ie daemon)
          match: "%{SYSLOGBASE}.*[Ee]rrors?(?!${WORD})"
          labels:
            program: "{{.program}}"
          path: /var/log/syslog
        """

        prometheusgrokexporter.charm_config["metrics"] = metrics_valid_yaml

        parsed_metrics = [
            {
                "type": "counter",
                "name": "syslog_msgs_errors_total",
                "help": "Total number of syslog msgs with errors, "
                "by program (ie daemon)",
                "match": "%{SYSLOGBASE}.*[Ee]rrors?(?!${WORD})",
                "labels": {"program": "{{.program}}", "host": hostname},
                "paths": ["/var/log/syslog"],
            }
        ]
        context = {
            "fail_on_missing_logfile": True,
            "global_paths": set(),
            "grok_patterns_path": "/var/snap/prometheus-grok-exporter"
            "/current/logstash-patterns-core/patterns",
            "metrics": parsed_metrics,
            "listen_ip": "0.0.0.0",
            "listen_port": 9144,
            "grok_patterns": {},
        }

        prometheusgrokexporter.write_grok_config()

        render.assert_called_with(
            source="grok-exporter.yml.j2",
            templates_dir="templates",
            target="/var/snap/prometheus-grok-exporter/current/etc/config.yml",
            context=context,
        )

    @mock.patch("lib_prometheus_grok_exporter.host.service_running")
    def test_is_service_running(self, service_running, prometheusgrokexporter):
        """Test is service running."""
        prometheusgrokexporter.is_service_running()
        service_running.assert_called_with(
            "snap.prometheus-grok-exporter.grok-exporter"
        )

    @mock.patch("lib_prometheus_grok_exporter.host.service_restart")
    @mock.patch("lib_prometheus_grok_exporter.host.service_start")
    def test_start_service(
        self, service_start, service_restart, prometheusgrokexporter
    ):
        """Test starting snap.prometheus-grok-exporter.grok-exporter service."""
        prometheusgrokexporter.start_service()
        service_start.assert_called_with("snap.prometheus-grok-exporter.grok-exporter")
        service_restart.assert_not_called()

    @mock.patch("lib_prometheus_grok_exporter.host.service_restart")
    @mock.patch("lib_prometheus_grok_exporter.host.service_start")
    def test_restart_service(
        self, service_start, service_restart, prometheusgrokexporter
    ):
        """Test restarting snap.prometheus-grok-exporter.grok-exporter service."""
        prometheusgrokexporter.start_service(restart=True)
        service_restart.assert_called_with(
            "snap.prometheus-grok-exporter.grok-exporter"
        )
        service_start.assert_not_called()
